//===----------------------------------------------------------------------===//
// INTEL CONFIDENTIAL
//
// Copyright 2022 Intel Corporation.
//
// This software and the related documents are Intel copyrighted materials, and
// your use of them is governed by the express license under which they were
// provided to you ("License"). Unless the License provides otherwise, you may
// not use, modify, copy, publish, distribute, disclose or transmit this
// software or the related documents without Intel's prior written permission.
//
// This software and the related documents are provided as is, with no express
// or implied warranties, other than those that are expressly stated in the
// License.
//===----------------------------------------------------------------------===//
//
// VQE
// Variational Quantum Eigensolver
// H = 0.5 Z^Z + 0.5 X^I + 0.25 I^X
// Target value (from classical eigensolver) = -0.9013878188659974

#include <clang/Quantum/quintrinsics.h>
#include <quantum.hpp>

#include <dlib/optimization.h>

const int total_qubits = 2;
const int N_params = 4;

qbit QubitReg[total_qubits];

typedef dlib::matrix<double, N_params, 1> column_vector;

static int steps_count = 0;

//Quantum Kernels
quantum_kernel void prep_z_all() {
  for (int i = 0; i < total_qubits; i++) {
      PrepZ(QubitReg[i]);
  }
}

quantum_kernel void ansatz_linear(double QuantumVariableParams[]) {
  int params_idx = 0;

  for (int i = 0; i < total_qubits; i++) {
      RY(QubitReg[i], QuantumVariableParams[params_idx++]);
  }

  CNOT(QubitReg[0], QubitReg[1]);

  for (int i = 0; i < total_qubits; i++) {
      RY(QubitReg[i], QuantumVariableParams[params_idx++]);
  }
}

quantum_kernel void hadamard_all() {
  for (int i = 0; i < total_qubits; i++) {
      H(QubitReg[i]);
  }
}

quantum_kernel void vqe_Z(double QuantumVariableParams[]) {
  prep_z_all();

  ansatz_linear(QuantumVariableParams);
}

/*
ZZ = 
1   0    0   0
0  -1    0   0
0   0   -1   0
0   0    0   1
*/
double Z0Z1_expectation(std::vector<double> ProbReg) {
  auto Z0Z1 = ProbReg[0] - ProbReg[1] - ProbReg[2] + ProbReg[3];
  return Z0Z1;
}

/*
ZI = 
1   0    0   0
0   1    0   0
0   0   -1   0
0   0    0  -1
*/
double Z0_expectation(std::vector<double> ProbReg) {
  auto Z0 = ProbReg[0] + ProbReg[1] - ProbReg[2] - ProbReg[3];
  return Z0;
}

/*
IZ = 
1   0    0   0
0  -1    0   0
0   0    1   0
0   0    0   1
*/
double Z1_expectation(std::vector<double> ProbReg) {
  auto Z1 = ProbReg[0] - ProbReg[1] + ProbReg[2] - ProbReg[3];
  return Z1;
}


int main() {
  iqsdk::IqsConfig settings(total_qubits, "noiseless");
  settings.verbose = true;
  iqsdk::FullStateSimulator quantum_8086(settings);

  if (iqsdk::QRT_ERROR_SUCCESS != quantum_8086.ready())
    return 1;

  
  std::vector<std::reference_wrapper<qbit>> qubit_ids;
  for (int id = 0; id < total_qubits; ++id) {
      qubit_ids.push_back(std::ref(QubitReg[id]));
  }

  double QuantumVariableParams[N_params];
  column_vector params_bobyqa = {0.1, 0.1, 0.1, 0.1};// Starting parameters

  auto EnergyOfAnsatz = [&](const column_vector& params) {
    for (int i = 0; i < N_params; i++) {
        QuantumVariableParams[i] = params(i);
    }

    vqe_Z(QuantumVariableParams);
    std::vector<double> ProbRegZ = quantum_8086.getProbabilities(qubit_ids);

    hadamard_all();
    std::vector<double> ProbRegX = quantum_8086.getProbabilities(qubit_ids);

    double energy_Z0Z1 = Z0Z1_expectation(ProbRegZ);
    double energy_X0 = Z0_expectation(ProbRegX);
    double energy_X1 = Z1_expectation(ProbRegX);
    double total_energy = 0.5 * energy_Z0Z1 + 0.5 * energy_X0
                         + 0.25 * energy_X1;

    steps_count++;
    return total_energy;
  };

  // initialize the optimization algorithm
  auto opt_bobyqa_val = dlib::find_min_bobyqa(
        EnergyOfAnsatz,
        params_bobyqa, 
        2 * 4 + 1, // number of interpolation points
        dlib::uniform_matrix<double>(4, 1, -7.0), // lower bound constraint
        dlib::uniform_matrix<double>(4, 1, 7.0), // upper bound constraint
        1.5, // initial trust region radius
        1e-5, // stopping trust region radius
        10000 // max number of objective function evaluations
        );

  std::cout.precision(17);
  std::cout << "Optimized Parameters using BOBYQA: "
            << "\n" << params_bobyqa;
  std::cout << "BOBYQA "
                "execution count: "
            << steps_count << "\n";
  std::cout << "Total energy using BOBYQA: "
            << opt_bobyqa_val << "\n";

  return 0;
}
